@@ -1,15 +1,16 @@
 import dask.array as da
 import matplotlib
+import matplotlib.pyplot as plt
 import numpy as np
 import pandas as pd
 import pytest
 import scanpy as sc
-import spatialdata_plot  # noqa: F401
 from anndata import AnnData
 from spatial_image import to_spatial_image
-from spatialdata import SpatialData, get_element_instances
+from spatialdata import SpatialData, deepcopy, get_element_instances
 from spatialdata.models import TableModel
 
+import spatialdata_plot  # noqa: F401
 from tests.conftest import DPI, PlotTester, PlotTesterMeta
 
 RNG = np.random.default_rng(seed=42)
@@ -69,27 +70,95 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
     def test_plot_can_stack_render_labels(self, sdata_blobs: SpatialData):
         (
             sdata_blobs.pl.render_labels(
-                element="blobs_labels", na_color="red", fill_alpha=1, outline_alpha=0, outline=False
-            )
-            .pl.render_labels(
-                element="blobs_labels", na_color="blue", fill_alpha=0, outline_alpha=1, outline=True, contour_px=15
+                element="blobs_labels",
+                na_color="red",
+                fill_alpha=1,
+                outline_alpha=0,
             )
+            .pl.render_labels(element="blobs_labels", na_color="blue", fill_alpha=0, outline_alpha=1, contour_px=15)
             .pl.show()
         )
 
     def test_plot_can_color_labels_by_continuous_variable(self, sdata_blobs: SpatialData):
         sdata_blobs.pl.render_labels("blobs_labels", color="channel_0_sum").pl.show()
 
-    def test_plot_can_color_labels(self, sdata_blobs: SpatialData):
-        table = sdata_blobs["table"].copy()
+    @pytest.mark.parametrize(
+        "label",
+        [
+            "blobs_labels",
+            "blobs_multiscale_labels",
+        ],
+    )
+    def test_plot_can_color_labels_by_categorical_variable(self, sdata_blobs: SpatialData, label: str):
+
+        def _make_tablemodel_with_categorical_labels(sdata_blobs, label):
+
+            n_obs = len(get_element_instances(sdata_blobs[label]))
+            vals = np.arange(n_obs) + 1
+            adata = AnnData(vals.reshape(-1, 1), obs=pd.DataFrame({"instance_id": vals}))
+            adata.obs["category"] = pd.Categorical(
+                list(["a", "b", "c"] * ((n_obs // 3) + 1))[:n_obs],
+                categories=["a", "b", "c"],
+                ordered=True,
+            )
+            adata.obs["region"] = label
+            table = TableModel.parse(
+                adata=adata,
+                region_key="region",
+                instance_key="instance_id",
+                region=label,
+            )
+            sdata_blobs.tables["other_table"] = table
+
+            sdata_blobs.pl.render_labels(label, color="category", table="other_table", scale="scale0").pl.show()
+
+        # we're modifying the data here, so we need an independent copy
+        sdata_blobs_local = deepcopy(sdata_blobs)
+
+        _make_tablemodel_with_categorical_labels(sdata_blobs_local, label)
+
+    def test_plot_two_calls_with_coloring_result_in_two_colorbars(self, sdata_blobs: SpatialData):
+        # we're modifying the data here so we need an independent copy
+        sdata_blobs_local = deepcopy(sdata_blobs)
+
+        table = sdata_blobs_local["table"].copy()
         table.obs["region"] = "blobs_multiscale_labels"
         table.uns["spatialdata_attrs"]["region"] = "blobs_multiscale_labels"
         table = table[:, ~table.var_names.isin(["channel_0_sum"])]
-        sdata_blobs["multi_table"] = table
-        sdata_blobs.pl.render_labels("blobs_labels", color="channel_0_sum", table_name="table").pl.render_labels(
+        sdata_blobs_local["multi_table"] = table
+        sdata_blobs_local.pl.render_labels("blobs_labels", color="channel_0_sum", table_name="table").pl.render_labels(
             "blobs_multiscale_labels", color="channel_1_sum", table_name="multi_table"
         ).pl.show()
 
+    def test_plot_can_control_label_outline(self, sdata_blobs: SpatialData):
+        sdata_blobs.pl.render_labels(
+            "blobs_labels", color="channel_0_sum", outline_alpha=0.4, fill_alpha=0.0, contour_px=15
+        ).pl.show()
+
+    def test_plot_can_control_label_infill(self, sdata_blobs: SpatialData):
+        sdata_blobs.pl.render_labels(
+            "blobs_labels",
+            color="channel_0_sum",
+            outline_alpha=0.0,
+            fill_alpha=0.4,
+        ).pl.show()
+
+    def test_plot_label_colorbar_uses_alpha_of_less_transparent_infill(
+        self,
+        sdata_blobs: SpatialData,
+    ):
+
+        sdata_blobs.pl.render_labels(
+            "blobs_labels", color="channel_0_sum", fill_alpha=0.1, outline_alpha=0.7, contour_px=15
+        ).pl.show()
+
+    def test_plot_label_colorbar_uses_alpha_of_less_transparent_outline(
+        self,
+        sdata_blobs: SpatialData,
+    ):
+
+        sdata_blobs.pl.render_labels("blobs_labels", color="channel_0_sum", fill_alpha=0.7, outline_alpha=0.1).pl.show()
+
     def test_can_plot_with_one_element_color_table(self, sdata_blobs: SpatialData):
         table = sdata_blobs["table"].copy()
         table.obs["region"] = "blobs_multiscale_labels"
@@ -100,32 +169,28 @@ class TestLabels(PlotTester, metaclass=PlotTesterMeta):
             "blobs_multiscale_labels", color="channel_1_sum", table_name="multi_table"
         ).pl.show()
 
-    @pytest.mark.parametrize(
-        "label",
-        [
+    def test_plot_subset_categorical_label_maintains_order(self, sdata_blobs: SpatialData):
+        max_col = sdata_blobs.table.to_df().idxmax(axis=1)
+        max_col = pd.Categorical(max_col, categories=sdata_blobs.table.to_df().columns, ordered=True)
+        sdata_blobs.table.obs["which_max"] = max_col
+
+        _, axs = plt.subplots(nrows=1, ncols=2, layout="tight")
+
+        sdata_blobs.pl.render_labels("blobs_labels", color="which_max").pl.show(ax=axs[0])
+        sdata_blobs.pl.render_labels(
             "blobs_labels",
-            "blobs_multiscale_labels",
-        ],
-    )
-    def test_plot_label_categorical_color(self, sdata_blobs: SpatialData, label: str):
-        self._make_tablemodel_with_categorical_labels(sdata_blobs, label)
+            color="which_max",
+            groups=["channel_0_sum"],
+        ).pl.show(ax=axs[1])
 
-    def _make_tablemodel_with_categorical_labels(self, sdata_blobs, label):
+    def test_plot_subset_categorical_label_maintains_order_when_palette_overwrite(self, sdata_blobs: SpatialData):
+        max_col = sdata_blobs.table.to_df().idxmax(axis=1)
+        max_col = pd.Categorical(max_col, categories=sdata_blobs.table.to_df().columns, ordered=True)
+        sdata_blobs.table.obs["which_max"] = max_col
 
-        n_obs = max(get_element_instances(sdata_blobs[label]))
-        vals = np.arange(n_obs)
-        obs = pd.DataFrame({"a": vals, "b": vals + 0.3, "c": vals + 0.7})
+        _, axs = plt.subplots(nrows=1, ncols=2, layout="tight")
 
-        adata = AnnData(vals.reshape(-1, 1), obs=obs)
-        adata.obs["instance_id"] = vals
-        adata.obs["category"] = list(["a", "b", "c"] * ((n_obs // 3) + 1))[:n_obs]
-        adata.obs["region"] = label
-        table = TableModel.parse(
-            adata=adata,
-            region_key="region",
-            instance_key="instance_id",
-            region=label,
-        )
-        sdata_blobs["other_table"] = table
-        sdata_blobs["other_table"].obs["category"] = sdata_blobs["other_table"].obs["category"].astype("category")
-        sdata_blobs.pl.render_labels(label, color="category").pl.show()
+        sdata_blobs.pl.render_labels("blobs_labels", color="which_max").pl.show(ax=axs[0])
+        sdata_blobs.pl.render_labels(
+            "blobs_labels", color="which_max", groups=["channel_0_sum"], palette="red"
+        ).pl.show(ax=axs[1])
