@@ -2,11 +2,11 @@ import matplotlib
 import numpy as np
 import pandas as pd
 import scanpy as sc
-import spatialdata_plot  # noqa: F401
 from anndata import AnnData
 from spatialdata import SpatialData
 from spatialdata.models import TableModel
 
+import spatialdata_plot  # noqa: F401
 from tests.conftest import DPI, PlotTester, PlotTesterMeta
 
 RNG = np.random.default_rng(seed=42)
@@ -30,7 +30,7 @@ class TestPoints(PlotTester, metaclass=PlotTesterMeta):
     def test_plot_can_filter_with_groups(self, sdata_blobs: SpatialData):
         sdata_blobs["table"].obs["region"] = ["blobs_points"] * sdata_blobs["table"].n_obs
         sdata_blobs["table"].uns["spatialdata_attrs"]["region"] = "blobs_points"
-        sdata_blobs.pl.render_points(color="genes", groups="gene_b", palette="orange").pl.show()
+        sdata_blobs.pl.render_points(color="genes", groups="gene_b", palette="red").pl.show()
 
     def test_plot_can_filter_with_groups_default_palette(self, sdata_blobs: SpatialData):
         sdata_blobs["table"].obs["region"] = ["blobs_points"] * sdata_blobs["table"].n_obs
